@@ -38,6 +38,8 @@ class Connection(object):
 @attrs(eq=False)
 class Configuration(object):
 
+    parallelDoStep = attrib(type=bool, default=False, repr=False)
+
     description = attrib(type=str, default=None, repr=False)
     variableNamingConvention = attrib(type=str, default='flat', repr=False)
 
@@ -93,6 +95,7 @@ def create_fmu_container(configuration, output_filename):
     os.mkdir(os.path.join(unzipdir, 'resources'))
 
     data = {
+        'parallelDoStep': configuration.parallelDoStep,
         'components': [],
         'variables': [],
         'connections': []
@@ -182,10 +185,22 @@ def create_fmu_container(configuration, output_filename):
             component_indices.append(component_index)
             value_references.append(value_reference)
 
-        data['variables'].append({
+        variable = {
             'components': component_indices,
-            'valueReferences': value_references
-        })
+            'valueReferences': value_references,
+        }
+
+        if v.start is not None:
+            if v.type == 'Real':
+                variable['start'] = float(v.start)
+            elif v.type in ['Enumeration', 'Integer']:
+                variable['start'] = int(v.start)
+            elif v.type == 'Boolean':
+                variable['start'] = bool(v.start)
+            elif v.type == 'String':
+                variable['start'] = v.start
+
+        data['variables'].append(variable)
 
         # modelDescription.xml
         start = f' start="{ v.start }"' if v.start else ''
